"""
Enumeration of patching strategies
"""

from enum import Enum


class STRATEGY(str, Enum):
    """
    Patching strategy enumeration
    """

    CAVE = "CAVE"
    ADD = "ADD"
    REPLACE = "REPLACE"



"""
Binary program wrapper
"""

from dataclasses import dataclass
from io import BytesIO
from logging import getLogger
from pathlib import Path
from typing import Callable, Dict, List, Optional, Union, cast

from angr import Project
from cle.backends import Backend
from lief import Binary as LIEFBinary  # pylint: disable=no-name-in-module
from lief import parse  # pylint: disable=no-name-in-module
from lief.ELF import (  # pylint: disable=no-name-in-module,import-error
    SEGMENT_FLAGS,
    SEGMENT_TYPES,
    Segment,
)
from capstone import CsInsn
from pysquishy.squishy import Squishy

from pypatches.error import NoSectionError
from pypatches.types import Code, TransformInfo

logger = getLogger(__name__)


@dataclass
class WriteOperation:
    """
    A write operation to be performed on a section

    :attribute data: Either raw bytes or a function that takes the resolved
        labels for all patches and returns raw bytes

    :attribute vaddr: The virtual address or label location to write to
    """

    data: Union[bytes, Callable[[Dict[str, int]], bytes], Code]
    where: Union[str, int]


class BinaryManager:
    """
    Wrapper for a binary program to provide information for patching
    """

    path: Optional[Path] = None
    blob: BytesIO
    lief_binary: LIEFBinary
    angr_project: Project
    cle_binary: Backend
    cle_opts: Dict[str, bool] = {
        "auto_load_libs": False,
        "use_system_libs": False,
    }
    cfg_opts: Dict[str, bool] = {
        "normalize": True,
        "data_references": False,
        "cross_references": False,
        "skip_unmapped_addrs": True,
        "force_complete_scan": False,
    }
    writes: List[WriteOperation] = []
    code_to_add: Dict[str, Code] = {}

    def __init__(
        self,
        binary: Union[Path, str, bytes],
        cle_opts: Optional[Dict[str, bool]] = None,
        cfg_opts: Optional[Dict[str, bool]] = None,
        silence_angr_logs: bool = False,
    ) -> None:
        """
        Initialize the binary wrapper via one of several methods.

        :param binary: If `binary` is a `Path` or `str` we will attempt
            to load it from that path on disk. If it is `bytes`, we will
            first try to load it from the decoded path, and if there is
            no such path, we will try to load it as a binary blob.
        :param cle_opts: If provided, will override the kwargs passed to
            `cle.Loader`
        """
        if isinstance(binary, Path):
            self.path = binary
        elif isinstance(binary, str):
            self.path = Path(binary)
        elif isinstance(binary, bytes):
            self.path = None
        else:
            raise TypeError(
                f"Requested binary is of type {type(binary)}"
                ", expected Path, str, or bytes."
            )

        if self.path is not None and not self.path.is_file():
            raise FileNotFoundError(
                f"Requested binary {self.path} was not found or could not be opened."
            )

        if cle_opts is not None:
            self.cle_opts = cle_opts

        if cfg_opts is not None:
            self.cfg_opts = cfg_opts

        if silence_angr_logs:
            for logger_name in ("angr", "pyvex", "cle", "archinfo", "claripy"):
                getLogger(logger_name).setLevel("ERROR")

        if self.path is None:
            self.blob = BytesIO(cast(bytes, binary))
            self.lief_binary = parse(binary)
            self.angr_project = Project(
                self.blob,
                main_opts={"base_addr": self.lief_binary.imagebase},
                load_options=self.cle_opts,
            )
            self.angr_project.analyses.CFGFast(  # type: ignore
                **self.cfg_opts,
            )

            if self.angr_project.loader.main_object is None:
                raise FileNotFoundError(
                    f"Requested binary {self.path} "
                    "was not found or could not be opened."
                )

            self.cle_binary = self.angr_project.loader.main_object
        else:
            self.blob = BytesIO(self.path.read_bytes())
            self.lief_binary = parse(str(self.path))
            self.angr_project = Project(
                str(self.path),
                main_opts={"base_addr": self.lief_binary.imagebase},
                load_options=self.cle_opts,
            )
            self.angr_project.analyses.CFGFast(  # type: ignore
                **self.cfg_opts,
            )

            if self.angr_project.loader.main_object is None:
                raise FileNotFoundError(
                    f"Requested binary {self.path} "
                    "was not found or could not be opened."
                )

            self.cle_binary = self.angr_project.loader.main_object

    def write(
        self,
        where: Union[int, str],
        data: Union[bytes, Callable[[Dict[str, int]], bytes], Code],
    ) -> None:
        """
        Write data to the binary at the given virtual address

        :param where: Either an address or a label that will resolve to an address
        :param data: The bytes or a function that takes a dictionary of labels
            and addresses and returns bytes to write to the patch
        """

        operation = WriteOperation(data, where)
        logger.info(f"Queuing write operation: {operation}")
        self.writes.append(operation)

    def read(self, vaddr: int, size: int) -> bytes:
        """
        Read data from the binary at the given virtual address
        """
        try:
            section = next(
                filter(lambda s: s.contains_addr(vaddr), self.cle_binary.sections)
            )
        except StopIteration as e:
            logger.error(f"Could not find section for address {vaddr}")
            raise NoSectionError(f"Could not find section for address {vaddr}") from e

        self.blob.seek(section.addr_to_offset(vaddr))
        return self.blob.read(size)

    def code(self, code: Code) -> bytes:
        """
        Get the raw code from a code object

        :param code: The code object
        """
        if code.c_code:
            compiled_code = Squishy().compile(code.c_code)

            disassembly = self.angr_project.arch.disasm(compiled_code, 0)
            logger.info(f"Disassembly: {disassembly}")

            return compiled_code

        if code.assembly:
            # TODO: Handle PC-relative assembly by making this a callable taking the addr
            return self.asm(code.assembly, 0)

        if code.raw:
            return code.raw

        raise ValueError("Code object has no code or assembly")

    def add_code(self, code: Code, label: Optional[str] = None) -> None:
        """
        Mark some code as being added on next save

        :param code: The binary code being added
        :param label: The label to associate with the code for resolution of
            other patches referencing it
        """
        if label is None:
            label = f"code_{len(self.code_to_add)}"

        logger.info(f"Queueing code addition at label {label}: {code}")

        self.code_to_add[label] = code

    def save(self, where: Path) -> None:
        """
        Save the binary to where

        :param where: Path to save the binary to
        """
        offsets = {}
        base_addr = None

        text_section_offset = self.lief_binary.get_section(".text").offset

        # Add code to the binary if there is any to add
        if self.code_to_add:
            segment = Segment()
            content = b""

            for label, code in self.code_to_add.items():
                logger.info(f"Adding code at label {label}: {code}")
                offsets[label] = len(content)
                content += self.code(code)

            segment.content = list(content)
            segment.type = SEGMENT_TYPES.LOAD
            segment.alignment = self.cle_binary.arch.instruction_alignment
            segment.flags = SEGMENT_FLAGS(5)
            new_segment = self.lief_binary.add(segment)
            base_addr = new_segment.virtual_address

            for label, offset in offsets.items():
                offsets[label] += base_addr

        new_text_section_offset = self.lief_binary.get_section(".text").offset
        text_section_adjust = new_text_section_offset - text_section_offset

        # After adding code, the program headers move to the end of the file and
        # the .text section address changes on disk, so we need to perform a relocation
        # based on the new offset of the text section

        tinfo = TransformInfo(offsets, text_section_adjust, self.lief_binary)

        for write in self.writes:
            if isinstance(write.where, str):
                offset = offsets[write.where]
            else:
                offset = write.where

            offset += text_section_adjust

            if isinstance(write.data, bytes):
                data = write.data
            elif isinstance(write.data, Code):
                write.data.transform(tinfo)
                data = self.code(write.data)
            else:
                data = write.data(offsets)

            logger.info(f"Writing {len(data)} bytes at {offset:#0x}")

            self.lief_binary.patch_address(offset, list(data))

        self.lief_binary.write(str(where))

    def asm(self, asm: str, vaddr: int) -> bytes:
        """
        Assemble the given assembly code at the given virtual address
        """
        return self.cle_binary.arch.asm(asm, vaddr, as_bytes=True)  # type: ignore

    def disasm(self, vaddr: int) -> CsInsn:
        """
        Disassemble the binary at an address

        :param vaddr: The virtual address to disassemble at
        """
        block = (
            self.binary.angr_project.kb.cfgs["CFGFast"]
            .get_any_node(vaddr, anyaddr=True)
            .block
        )

        for instr in block.capstone.insns:
            if instr.address == vaddr:
                return instr

        raise ValueError(f"Could not find instruction at {vaddr:#0x}")

    def relative_call_addr(self, source: int, dest: int) -> int:
        """
        Calculate the relative offset of dest from source

        :param source: The address of the instruction making the call
        :param dest: The address of the instruction that should be called
            to
        """
        isize = self.disasm(source).size



"""
Patch descriptors for various types of patches

Patches require various parameters to specify how to apply them, but some are common.

`label`, if available, allows a label to be specified. In any later patch, that label
can be used to reference the location of that patch (this is especially useful for
data patches)
"""

from dataclasses import InitVar, dataclass, field
from typing import List, Optional, Set, Union

from pypatches.types import AddressRange, Code, ReturnValue, Value


@dataclass
class NopPatch:
    """
    Patch that converts specific addresses or ranges of addresses to no-operations
    """

    addresses: InitVar[Optional[List[int]]] = None
    address_ranges: Set[AddressRange] = field(default_factory=set)

    def __post_init__(self, addresses: Optional[List[int]]) -> None:
        """
        Convert optionally provided addresses to address ranges

        :param addresses: Addresses that can be provided in lieu of address ranges
        """
        if addresses is None:
            return

        for address in addresses:
            self.address_ranges.add(AddressRange(start=address, end=address))


@dataclass
class BranchPatch:
    """
    Base patch that modifies a conditional branch at a particular address
    """

    address: int


@dataclass
class InvertBranchPatch(BranchPatch):
    """
    Patch that inverts the true/false branch targets of a conditional branch
    """


@dataclass
class AlwaysBranchPatch(BranchPatch):
    """
    Patch that converts a conditional branch into an unconditional branch
    that always takes the "true" branch
    """


@dataclass
class NeverBranchPatch(BranchPatch):
    """
    Patch that converts a conditional branch into an unconditional branch
    that always takes the "false" branch
    """


@dataclass
class SkipAndReturnPatch:
    """
    Patch that skips a call to a subroutine at a particular address and
    fakes a return value as if the called function had returned it using
    the default calling convention
    """

    caller_address: int
    return_value: ReturnValue


@dataclass
class FunctionReplacePatch:
    """
    Patch that will replace a function's contents with new code
    """

    function_address: int
    new_code: Code


@dataclass
class CallerReplacePatch:
    """
    Patch that will redirect all or some callers of a function elsewhere
    """

    new_code: Code
    function_address: Optional[int] = None
    callers: Set[int] = field(default_factory=set)

    def __post_init__(self) -> None:
        """
        Check that we either got a function address or callers, or both
        """
        if self.function_address is None and not self.callers:
            raise ValueError(
                "A function address and/or set of callers must be provided."
            )


@dataclass
class InitPatch:
    """
    Patch that will run some code immediately on entry to the program
    """

    priority: int
    code: Code


@dataclass
class FiniPatch:
    """
    Patch that will run some code upon exit from the program
    """

    priority: int
    code: Code


@dataclass
class DataPatch:
    """
    Patch that adds some data with some protections
    """

    data: Value
    read: bool = True
    write: bool = False
    exec: bool = False
    label: Optional[str] = None


@dataclass
class CodePatch:
    """
    Patch that holds some code
    """

    code: Code


@dataclass
class AddCodePatch(CodePatch):
    """
    Patch that adds some code to the binary at some location
    """

    label: Optional[str] = None


@dataclass
class ReplaceCodePatch(CodePatch):
    """
    Patch that replaces some code with some other code
    """

    address: int


PatchType = Union[
    NopPatch,
    BranchPatch,
    InvertBranchPatch,
    AlwaysBranchPatch,
    NeverBranchPatch,
    SkipAndReturnPatch,
    FunctionReplacePatch,
    CallerReplacePatch,
    InitPatch,
    DataPatch,
    AddCodePatch,
    ReplaceCodePatch,
]



"""
Errors specific to the patches package.
"""


class CompilationError(Exception):
    """
    An error that occurs when a snippet of C/C++ code fails to compile
    """


class TransformationError(Exception):
    """
    An error that occurs when transformation of LLVM bitcode using the
    wrapper fails
    """


class CodegenError(Exception):
    """
    An error that occurs when code generation fails
    """


class BinaryCreateError(Exception):
    """
    An error that occurs when a binary code file cannot be produced
    """


class NoSectionError(Exception):
    """
    An error that occurs when a section is not found
    """



"""
Common types used by patches
"""

from dataclasses import dataclass, field
from enum import Enum
from typing import Callable, Dict, List, Optional, Tuple

from lief import Binary


class TYPECLASS(str, Enum):
    """
    An enum representing various classes a type may be a member of
    """

    BASIC = "BASIC"
    POINTER = "POINTER"
    STRUCT = "STRUCT"
    UNION = "UNION"
    FUNCTION = "FUNCTION"


@dataclass
class Type:
    """
    Type representing a Type in a (C/ASM) type language
    """

    label: str
    size: int
    signed: bool = False
    typeclass: TYPECLASS = TYPECLASS.BASIC
    fields: List[Tuple[Optional[str], "Type"]] = field(default_factory=list)
    base: Optional["Type"] = None


@dataclass
class Value(Type):
    """
    Value of a given type, where the value is None if the type is a pointer
    """

    value: Optional[bytes] = None


@dataclass
class ReturnValue(Value):
    """
    A return value, which can be of any `Type` and its
    value
    """


@dataclass(frozen=True)
class AddressRange:
    """
    An address range with a start and end.
    If start == end, this address range is just one address
    """

    start: int
    end: int

    @property
    def address(self) -> bool:
        """
        Returns whether this range is "an address"
        """
        return self.start == self.end


@dataclass
class TransformInfo:
    """
    Information container used for transforming code based on context

    :attribute label_offsets: Dictionary of label: address of the addresses
        of each label added to the program
    :attribute text_offset: If the text section has moved, add this to any
        address that was in the text section to refer to it and reflect
        the movement
    """

    label_offsets: Dict[str, int]
    text_offset: int
    lief_binary: Binary


@dataclass
class Code:
    """
    Code either in assembly, C, or already assembled bytes
    """

    assembly: Optional[str] = None
    c_code: Optional[str] = None
    raw: Optional[bytes] = None
    transform_asm: Optional[Callable[[str, TransformInfo], str]] = None
    transform_c_code: Optional[Callable[[str, TransformInfo], str]] = None
    transform_raw: Optional[Callable[[bytes, TransformInfo], bytes]] = None

    def __post_init__(self) -> None:
        """
        Check that we got one of the above
        """
        if self.assembly is None and self.c_code is None and self.raw is None:
            raise ValueError("No code was provided.")

    def transform(self, tinfo: TransformInfo) -> None:
        """
        Call a transformer function if one is present that modifies the current code
        using the label to address mapping provided

        :param offsets: A mapping of labels to addresses
        """
        if self.assembly is not None and self.transform_asm is not None:
            self.assembly = self.transform_asm(self.assembly, tinfo)
        if self.c_code is not None and self.transform_c_code is not None:
            self.c_code = self.transform_c_code(self.c_code, tinfo)
        if self.raw is not None and self.transform_raw is not None:
            self.raw = self.transform_raw(self.raw, tinfo)

    @classmethod
    def build_c_code(
        cls,
        body: str,
        getreg_helper: bool = True,
        includes: Optional[List[str]] = None,
        extra_code: Optional[str] = None,
    ) -> str:
        """
        Build a C code snippet from a body of code.

        :param body: The body of the code to build, which will be wrapped in
            a main function and compiled.
        :param getreg_helper: Whether to include the getreg helper macro that
            can be used by calling `getreg(variable_name, reg_name)` in the body
            of the function
        :param includes: A list of include files like `["#include <stdint.h>", ...]`
        """
        code = ""

        if includes is not None:
            code = "\n".join(includes)
            code += "\n"

        if extra_code is not None:
            code += extra_code + "\n"

        if getreg_helper:
            code += (
                """#define getreg(dest, src)  \\\n"""
                """    register long long dest __asm__ (#src); \\\n"""
                """    __asm__ ("" :"=r"(dest));\n"""
            )

        code += """int main() {\n"""
        code += body + "\n"
        code += "}"

        return code



"""
Check if two capstone MemOp objects are equal, since they don't
have a __eq__ method that works.
"""

from ctypes import Structure
from typing import Tuple

from capstone.arm import ArmOpMem
from capstone.arm64 import Arm64OpMem
from capstone.m68k import M68KOpMem
from capstone.mips import MipsOpMem
from capstone.ppc import PpcOpMem
from capstone.sparc import SparcOpMem
from capstone.systemz import SyszOpMem
from capstone.tms320c64x import TMS320C64xOpMem
from capstone.x86 import X86OpMem
from capstone.xcore import XcoreOpMem


def cs_memop_eq(first: Structure, second: Structure) -> bool:
    """
    Check equality of two memory operands

    :param first: The first operand
    :param second: The second operand
    """
    # If they aren't the same subclass, they can't possibly be equal
    # in a memory operand sense
    if first.__class__.__name__ != second.__class__.__name__:
        return False

    check: Tuple[str, ...] = ()
    if isinstance(first, ArmOpMem):
        check = ("base", "index", "scale", "disp", "lshift")
    elif isinstance(first, M68KOpMem):
        check = (
            "base_reg",
            "index_reg",
            "in_base_reg",
            "in_disp",
            "out_disp",
            "disp",
            "scale",
        )
    elif isinstance(first, MipsOpMem):
        check = ("base", "disp")
    elif isinstance(first, Arm64OpMem):
        check = ("base", "index", "disp")
    elif isinstance(first, PpcOpMem):
        check = ("base", "disp")
    elif isinstance(first, X86OpMem):
        check = ("segment", "base", "index", "scale", "disp")
    elif isinstance(first, XcoreOpMem):
        check = ("base", "index", "disp", "direct")
    elif isinstance(first, TMS320C64xOpMem):
        check = ("base", "disp", "unit", "scaled", "disptype", "direction", "modify")
    elif isinstance(first, SparcOpMem):
        check = ("base", "index", "disp")
    elif isinstance(first, SyszOpMem):
        check = ("base", "index", "length", "disp")

    for attr in check:
        if getattr(first, attr) != getattr(second, attr):
            return False

    return True
